@@ -9,14 +9,15 @@ def _assemble_intensity_and_fluxes(
     scaled_omega_arr,                   # Delta-scaled single-scattering albedos
     tau_arr,                            # Lower boundary of layers
     scaled_tau_arr_with_0,              # Delta-scaled lower boundary of layers with 0 inserted in front
-    mu_arr_pos, mu_arr,                 # Quadrature nodes for 1) upper 2) both hemispheres
-    M_inv, W,                           # 1) 1 / mu; 2) quadrature weights for each hemisphere
+    mu_arr_pos,                         # Upper hemisphere quadrature nodes (cosine polar angles)
+    M_inv, W,                           # 1) 1 / `mu_arr_pos`; 2) quadrature weights for each hemisphere
     N, NQuad, NLeg,                     # Number of 1) upper 2) both hemispheres quadrature nodes; 3) phase function Legendre coefficients 
     NFourier, NLayers, NBDRF,           # Number of 1) intensity Fourier modes; 2) layers; 3) BDRF Fourier modes
     is_atmos_multilayered,              # Is the atmosphere multilayered?
     weighted_scaled_Leg_coeffs,         # Weighted and delta-scaled Legendre coefficients
     BDRF_Fourier_modes,                 # BDRF Fourier modes
-    mu0, I0, rescale_factor, phi0,      # Properties of the direct beam
+    mu0, I0, I0_div_4pi,                # Primary direct beam parameters
+    rescale_factor, phi0,               # Secondary direct beam parameters
     there_is_beam_source,               # Is there a beam source?
     b_pos, b_neg,                       # Dirichlet BCs
     b_pos_is_scalar, b_neg_is_scalar,   # Is each Dirichlet BCs scalar and so isotropic?
@@ -47,7 +48,6 @@ def _assemble_intensity_and_fluxes(
     | `tau_arr`                      | `NLayers`                                    |
     | `scaled_tau_arr_with_0`        | `NLayers + 1`                                |
     | `mu_arr_pos`                   | `NQuad/2`                                    |
-    | `mu_arr`                       | `NQuad`                                      |
     | `M_inv`                        | `NQuad/2`                                    |
     | `W`                            | `NQuad/2`                                    |
     | `N`                            | scalar                                       |
@@ -61,6 +61,7 @@ def _assemble_intensity_and_fluxes(
     | `BDRF_Fourier_modes`           | `NBDRF`                                      |
     | `mu0`                          | scalar                                       |
     | `I0`                           | scalar                                       |
+    | `I0_div_4pi`                   | scalar                                       |
     | `rescale_factor`               | scalar                                       |
     | `phi0`                         | scalar                                       |
     | `there_is_beam_source`         | scalar                                       |
@@ -83,7 +84,7 @@ def _assemble_intensity_and_fluxes(
     | ----------------- | -------------------------------------- |
     | `G_collect`       | `NFourier x NLayers x NQuad x NQuad`   |
     | `G_collect_0`     | `NLayers x NQuad x NQuad`              |
-    | `G_inv_collect_0` | `NLayers x NQuad x NQuad` or `None`    |
+    | `G_inv_mu_inv`    | `NLayers x NQuad x NQuad` or `None`    |
     | `K_collect`       | `NFourier x NLayers x NQuad`           |
     | `K_collect_0`     | `NLayers x NQuad`                      |
     | `B_collect`       | `NFourier x NLayers x NQuad` or `None` |
@@ -108,45 +109,45 @@ def _assemble_intensity_and_fluxes(
     outputs = _solve_for_gen_and_part_sols(
         NFourier,
         scaled_omega_arr,
-        mu_arr_pos, mu_arr,
+        mu_arr_pos,
         M_inv, W,
         N, NQuad, NLeg,
         NLayers,
         weighted_scaled_Leg_coeffs,
-        mu0, I0,
+        mu0, I0_div_4pi,
         there_is_beam_source,
-        Nscoeffs,
         there_is_iso_source,
     )
     if there_is_beam_source and there_is_iso_source:
         G_collect, K_collect, B_collect, G_inv_collect_0 = outputs
         B_collect_0 = B_collect[0, :, :]
+        G_inv_mu_inv = G_inv_collect_0 @ np.concatenate([M_inv, -M_inv])
     elif there_is_beam_source and not there_is_iso_source:
         G_collect, K_collect, B_collect = outputs
         B_collect_0 = B_collect[0, :, :]
-        G_inv_collect_0 = None
+        G_inv_mu_inv = None
     elif not there_is_beam_source and there_is_iso_source:
         G_collect, K_collect, G_inv_collect_0 = outputs
         B_collect = None
+        G_inv_mu_inv = G_inv_collect_0 @ np.concatenate([M_inv, -M_inv])
     else:
         G_collect, K_collect = outputs
         B_collect = None
-        G_inv_collect_0 = None
-        
+        G_inv_mu_inv = None
             
     GC_collect = _solve_for_coeffs(
         NFourier,
         G_collect,
         K_collect,
         B_collect,
-        G_inv_collect_0,
+        G_inv_mu_inv,
         scaled_tau_arr_with_0,
-        mu_arr, mu_arr_pos, mu_arr_pos * W,
+        mu_arr_pos, mu_arr_pos * W,
         N, NQuad,
         NLayers, NBDRF,
         is_atmos_multilayered,
         BDRF_Fourier_modes,
-        mu0, I0,
+        mu0, I0_div_4pi,
         there_is_beam_source,
         b_pos, b_neg,
         b_pos_is_scalar, b_neg_is_scalar,
@@ -166,7 +167,7 @@ def _assemble_intensity_and_fluxes(
  
         # Construct the intensity function (refer to section 3.7 of the Comprehensive Documentation)
         # --------------------------------------------------------------------------------------------------------------------------
-        def u(tau, phi, is_antiderivative_wrt_tau=False, return_Fourier_error=False, return_tau_arr=False):
+        def u(tau, phi, is_antiderivative_wrt_tau=False, return_Fourier_error=False, return_tau_arr=False, *, _return_l=False):
             """
             Intensity function with arguments `(tau, phi)` of types `(array or float, array or float)`.
             Returns an ndarray with axes corresponding to variation with `mu, tau, phi` respectively.
@@ -182,7 +183,7 @@ def _assemble_intensity_and_fluxes(
             
             # Atmospheric layer indices
             l = np.argmax(tau[:, None] <= tau_arr[None, :], axis=1)
-            scaled_tau_arr_l = scaled_tau_arr_with_0[l + 1]
+            scaled_tau_arr_l = scaled_tau_arr_with_0[1:][l]
             scaled_tau_arr_lm1 = scaled_tau_arr_with_0[l]
 
             # Delta-M scaling
@@ -204,13 +205,13 @@ def _assemble_intensity_and_fluxes(
             direct_beam_contribution = 0
             if is_antiderivative_wrt_tau:
                 if there_is_beam_source:
-                    direct_beam_contribution = (B_collect / (-scale_tau / mu0)[None, :, None])[
-                        :, l, :
-                    ] * np.exp(-scaled_tau / mu0)[None, :, None]
+                    direct_beam_contribution = (
+                        B_collect.transpose(0, 2, 1) / (-scale_tau / mu0)[None, None, :]
+                    )[:, :, l] * np.exp(-scaled_tau / mu0)[None, None, :]
 
                 um = (
                     np.einsum(
-                        "mtij, mtj -> mti",
+                        "mtij, mtj -> mit",
                         GC_collect[:, l, :, :],
                         np.exp(exponent) / (scale_tau[None, :, None] * K_collect)[:, l, :],
                         optimize=True,
@@ -220,12 +221,13 @@ def _assemble_intensity_and_fluxes(
             else:
                 if there_is_beam_source:
                     direct_beam_contribution = (
-                        B_collect[:, l, :] * np.exp(-scaled_tau / mu0)[None, :, None]
+                        B_collect.transpose(0, 2, 1)[:, :, l]
+                        * np.exp(-scaled_tau / mu0)[None, None, :]
                     )
 
                 um = (
                     np.einsum(
-                        "mtij, mtj -> mti", GC_collect[:, l, :, :], np.exp(exponent), optimize=True
+                        "mtij, mtj -> mit", GC_collect[:, l, :, :], np.exp(exponent), optimize=True
                     )
                     + direct_beam_contribution
                 )
@@ -234,32 +236,31 @@ def _assemble_intensity_and_fluxes(
             if there_is_iso_source:
                 l_uniq, l_inv = np.unique(l, return_inverse=True)
                 _mathscr_v_contribution = _mathscr_v(
-                    scaled_tau, l_inv, Nscoeffs,
+                    scaled_tau, l_inv, 
+                    Nscoeffs,
                     scaled_s_poly_coeffs[l_uniq],
                     G_collect_0[l_uniq],
                     K_collect_0[l_uniq],
-                    G_inv_collect_0[l_uniq],
-                    mu_arr,
+                    G_inv_mu_inv[l_uniq],
                     is_antiderivative_wrt_tau,
                     autograd_compatible,
                 )
                 
                 if NFourier == 1:
-                    um = um + _mathscr_v_contribution.T
+                    um = um + _mathscr_v_contribution
                 elif autograd_compatible:
-                    um = um + np.concatenate((_mathscr_v_contribution.T, np.zeros((NFourier - 1, len(tau), NQuad))))
+                    um = np.concatenate((um[[0], :, :] + _mathscr_v_contribution, um[1:, :, :]), axis=0)
                 else:
-                    um[0, :, :] += _mathscr_v_contribution.T
+                    um[0, :, :] += _mathscr_v_contribution
                 
-            u = np.einsum(
-                "mti, mp -> itp",
-                um,
-                np.cos(np.arange(NFourier)[:, None] * (phi0 - phi)[None, :]),
-                optimize=True,
+            u = np.sum(
+                um[:, :, :, None]
+                * np.cos(np.arange(NFourier)[:, None] * (phi0 - phi)[None, :])[:, None, None, :],
+                axis=0,
             )
             
             outputs = (rescale_factor * np.squeeze(u),)
-            outputs_len = 1
+            output_is_tuple = False
 
             if return_Fourier_error:
                 exponent = np.concatenate(
@@ -312,13 +313,20 @@ def _assemble_intensity_and_fluxes(
                 )
                 
                 outputs += (Fourier_error,)
-                outputs_len += 1
+                output_is_tuple = True
                 
             if return_tau_arr:
                 outputs += (tau_arr,)
-                outputs_len += 1
+                output_is_tuple = True
+            
+            if _return_l: # This indicates to the function that NT corrections are activated
+                output_is_tuple = True
+                outputs += (l,)
+                if there_is_iso_source:
+                    # The TMS correction will use these
+                    outputs += (l_uniq, l_inv)
                 
-            return outputs[0] if outputs_len == 1 else outputs
+            return outputs if output_is_tuple else outputs[0]
         # --------------------------------------------------------------------------------------------------------------------------
     
     # Construct u0
@@ -339,7 +347,7 @@ def _assemble_intensity_and_fluxes(
         
         # Atmospheric layer indices
         l = np.argmax(tau[:, None] <= tau_arr[None, :], axis=1)
-        scaled_tau_arr_l = scaled_tau_arr_with_0[l + 1]
+        scaled_tau_arr_l = scaled_tau_arr_with_0[1:][l]
         scaled_tau_arr_lm1 = scaled_tau_arr_with_0[l]
 
         # Delta-M scaling
@@ -403,12 +411,12 @@ def _assemble_intensity_and_fluxes(
         if there_is_iso_source:
             l_uniq, l_inv = np.unique(l, return_inverse=True)
             _mathscr_v_contribution = _mathscr_v(
-                scaled_tau, l_inv, Nscoeffs,
+                scaled_tau, l_inv,
+                Nscoeffs,
                 scaled_s_poly_coeffs[l_uniq],
                 G_collect_0[l_uniq],
                 K_collect_0[l_uniq],
-                G_inv_collect_0[l_uniq],
-                mu_arr,
+                G_inv_mu_inv[l_uniq],
                 is_antiderivative_wrt_tau,
                 autograd_compatible,
             )
@@ -448,7 +456,7 @@ def _assemble_intensity_and_fluxes(
         
         # Atmospheric layer indices
         l = np.argmax(tau[:, None] <= tau_arr[None, :], axis=1)
-        scaled_tau_arr_l = scaled_tau_arr_with_0[l + 1]
+        scaled_tau_arr_l = scaled_tau_arr_with_0[1:][l]
         scaled_tau_arr_lm1 = scaled_tau_arr_with_0[l]
 
         # Delta-M scaling
@@ -462,12 +470,12 @@ def _assemble_intensity_and_fluxes(
         if there_is_iso_source:
             l_uniq, l_inv = np.unique(l, return_inverse=True)
             _mathscr_v_contribution = _mathscr_v(
-                scaled_tau, l_inv, Nscoeffs,
+                scaled_tau, l_inv,
+                Nscoeffs,
                 scaled_s_poly_coeffs[l_uniq],
                 G_collect_0[l_uniq, :N, :],
                 K_collect_0[l_uniq],
-                G_inv_collect_0[l_uniq],
-                mu_arr,
+                G_inv_mu_inv[l_uniq],
                 is_antiderivative_wrt_tau,
                 autograd_compatible,
             )
@@ -530,7 +538,7 @@ def _assemble_intensity_and_fluxes(
 
         # Atmospheric layer indices
         l = np.argmax(tau[:, None] <= tau_arr[None, :], axis=1)
-        scaled_tau_arr_l = scaled_tau_arr_with_0[l + 1]
+        scaled_tau_arr_l = scaled_tau_arr_with_0[1:][l]
         scaled_tau_arr_lm1 = scaled_tau_arr_with_0[l]
 
         # Delta-M scaling
@@ -544,12 +552,12 @@ def _assemble_intensity_and_fluxes(
         if there_is_iso_source:
             l_uniq, l_inv = np.unique(l, return_inverse=True)
             _mathscr_v_contribution = _mathscr_v(
-                scaled_tau, l_inv, Nscoeffs,
+                scaled_tau, l_inv,
+                Nscoeffs,
                 scaled_s_poly_coeffs[l_uniq],
                 G_collect_0[l_uniq, N:, :],
                 K_collect_0[l_uniq],
-                G_inv_collect_0[l_uniq],
-                mu_arr,
+                G_inv_mu_inv[l_uniq],
                 is_antiderivative_wrt_tau,
                 autograd_compatible,
             )
