@@ -1,9 +1,8 @@
 from PythonicDISORT.subroutines import _mathscr_v
-from PythonicDISORT.subroutines import to_diag_ordered_form
 
 import numpy as np
+from numpy.lib.stride_tricks import as_strided
 import scipy as sc
-from math import pi
 
 
 def _solve_for_coeffs(
@@ -11,14 +10,14 @@ def _solve_for_coeffs(
     G_collect,                                # Eigenvector matrices
     K_collect,                                # Eigenvalues
     B_collect,                                # Coefficients vectors for particular solutions
-    G_inv_collect_0,                          # Inverse of eigenvector matrix for the 0th Fourier mode
+    G_inv_mu_inv,                             # Inverse of eigenvector matrix for the 0th Fourier mode
     scaled_tau_arr_with_0,                    # Delta-scaled lower boundary of layers with 0 inserted in front
-    mu_arr, mu_arr_pos, mu_arr_pos_times_W,   # Quadrature nodes for 1) both 2) upper hemispheres; 3) upper hemisphere quadrature nodes times weights
+    mu_arr_pos, mu_arr_pos_times_W,           # Upper hemisphere quadrature nodes (cosine polar angles); same but multiplied by weights
     N, NQuad,                                 # Number of 1) upper 2) both hemispheres quadrature nodes
     NLayers, NBDRF,                           # Number of 1) layers; 2) BDRF Fourier modes
     is_atmos_multilayered,                    # Is the atmosphere multilayered?
     BDRF_Fourier_modes,                       # BDRF Fourier modes
-    mu0, I0,                                  # Properties of the direct beam
+    mu0, I0_div_4pi,                          # Primary properties of the direct beam
     there_is_beam_source,                     # Is there a beam source?
     b_pos, b_neg,                             # Dirichlet BCs
     b_pos_is_scalar, b_neg_is_scalar,         # Is each Dirichlet BCs scalar and so isotropic?
@@ -44,9 +43,8 @@ def _solve_for_coeffs(
     | `G_collect`                    | `NFourier x NLayers x NQuad x NQuad`         |
     | `K_collect`                    | `NFourier x NLayers x NQuad`                 |
     | `B_collect`                    | `NFourier x NLayers x NQuad` or `None`       |
-    | `G_inv_collect_0`              | `NLayers x NQuad x NQuad` or `None`          |
+    | `G_inv_mu_inv`                 | `NLayers x NQuad x NQuad` or `None`          |
     | `scaled_tau_arr_with_0`        | `NLayers + 1`                                |
-    | `mu_arr`                       | `NQuad`                                      |
     | `mu_arr_pos`                   | `NQuad/2`                                    |
     | `mu_arr_pos_times_W`           | `NQuad/2`                                    |
     | `N`                            | scalar                                       |
@@ -56,7 +54,7 @@ def _solve_for_coeffs(
     | `is_atmos_multilayered`        | boolean                                      |
     | `BDRF_Fourier_modes`           | `NBDRF`                                      |
     | `mu0`                          | scalar                                       |
-    | `I0`                           | scalar                                       |
+    | `I0_div_4pi`                   | scalar                                       |
     | `there_is_beam_source`         | boolean                                      |
     | `b_pos`                        | `NQuad/2 x NFourier` or `NQuad/2` or scalar  |
     | `b_neg`                        | `NQuad/2 x NFourier` or `NQuad/2` or scalar  | 
@@ -78,11 +76,36 @@ def _solve_for_coeffs(
     ################################## Solve for coefficients of homogeneous solution ##########################################
     ############# Refer to section 3.6.2 (single-layer) and 4 (multi-layer) of the Comprehensive Documentation #################
     
-    dim = NLayers * NQuad
-    GC_collect = np.empty((NFourier, NLayers, NQuad, NQuad))
     use_banded_solver = (NLayers >= use_banded_solver_NLayers)
+    dim = NLayers * NQuad
+    C_collect = np.empty((NFourier, dim))
+    
+    RHS = np.empty(dim)
+    i_bot = dim - N # Equals `j_bot_right`
+    j_bot_left = dim - NQuad
+
     if use_banded_solver:
-        Nsupsubdiags = 3 * N - 1 # The number of super-diagonals equals the number of sub-diagonals
+        Nsupsubdiags = 3 * N - 1
+        LHS_dof = np.zeros((6 * N - 1, dim), order="F") # Diagonal ordered form
+        s0, s1 = LHS_dof.strides
+        col_stride = s1 - s0  # Stride when moving +1 col and -1 row
+        
+        def _view_slanted(i0, j0, nrows):
+            """
+            Return a view into `LHS_dof` using slanted mapping.
+            The function uses `numpy.lib.stride_tricks.as_strided` which is performant but dangerous, 
+            which is why this function is not exposed in `PythonicDISORT.subroutines`.
+            See https://numpy.org/devdocs/reference/generated/numpy.lib.stride_tricks.as_strided.html.
+            
+            Assume the following exist in the outer scope:
+                `LHS_dof`, `Nsupsubdiags`, `s0`, `col_stride`
+            """
+            r0 = Nsupsubdiags + i0 - j0
+            base = LHS_dof[r0 : r0 + nrows, j0 : j0 + N]
+            return as_strided(base, shape=(nrows, N), strides=(s0, col_stride))
+            
+    else:
+        LHS = np.zeros((dim, dim), order="F")
     
     # The following loops can easily be parallelized, but the speed-up is unlikely to be worth the overhead
     for m in range(NFourier):
@@ -91,8 +114,6 @@ def _solve_for_coeffs(
         
         G_collect_m = G_collect[m, :, :, :]
         K_collect_m = K_collect[m, :, :]
-        if there_is_beam_source:
-            B_collect_m = B_collect[m, :, :]
             
         # Generate mathscr_D and mathscr_X (BDRF terms)
         # Just for this part, refer to section 3.4.2 of the Comprehensive Documentation 
@@ -103,19 +124,21 @@ def _solve_for_coeffs(
                 mathscr_D_neg = (1 + m_equals_0 * 1) * BDRF_Fourier_modes_m
                 R = mathscr_D_neg * mu_arr_pos_times_W[None, :]
                 if there_is_beam_source:
-                    mathscr_X_pos = (mu0 * I0 / pi) * BDRF_Fourier_modes_m
+                    mathscr_X_pos = (mu0 * I0_div_4pi * 4) * BDRF_Fourier_modes_m
             else:
                 mathscr_D_neg = (1 + m_equals_0 * 1) * BDRF_Fourier_modes_m(mu_arr_pos, mu_arr_pos)
                 R = mathscr_D_neg * mu_arr_pos_times_W[None, :]
                 if there_is_beam_source:
-                    mathscr_X_pos = (mu0 * I0 / pi) * BDRF_Fourier_modes_m(
+                    mathscr_X_pos = (mu0 * I0_div_4pi * 4) * BDRF_Fourier_modes_m(
                         mu_arr_pos, np.array([mu0])
                     )[:, 0]
+                    
         # --------------------------------------------------------------------------------------------------------------------------
     
+    
         # Assemble RHS
         # --------------------------------------------------------------------------------------------------------------------------
-        # Ensure the BCs are of the correct shape
+        # Ensure the Dirchlet BCs are of the correct shape
         if b_pos_is_scalar and m_equals_0:
             b_pos_m = np.full(N, b_pos)
         elif b_pos_is_vector and m_equals_0:
@@ -134,32 +157,38 @@ def _solve_for_coeffs(
         else:
             b_neg_m = b_neg[:, m]
         
-        # _mathscr_v_contribution
+        # Top Dirichlet BC
+        RHS[:N] = b_neg_m
+        # Bottom Dirichlet BC
+        RHS[-N:] = b_pos_m
+        
+        # Reset the rest of `RHS` to 0
+        RHS[N:-N] = 0
+        
         if m_equals_0 and there_is_iso_source:
-            _mathscr_v_contribution_top = -_mathscr_v(
+            # Top boundary
+            RHS[:N] -= _mathscr_v(
                         np.array([0]), 
                         np.array([0]),
                         Nscoeffs,
                         scaled_s_poly_coeffs[[0], :],
                         G_collect_m[[0], N:, :],
                         K_collect_m[[0], :],
-                        G_inv_collect_0[[0], :, :],
-                        mu_arr,
+                        G_inv_mu_inv[[0]],
                     ).ravel()
         
-            _mathscr_v_contribution_middle = np.array([])
             if is_atmos_multilayered:
                 indices = np.arange(NLayers - 1)
-                _mathscr_v_contribution_middle = (
+                indicesp1 = indices + 1
+                RHS[N:-N] += (
                     _mathscr_v(
                         scaled_tau_arr_with_0[1:-1],
                         indices,
                         Nscoeffs,
-                        scaled_s_poly_coeffs[indices + 1],
-                        G_collect_m[indices + 1],
-                        K_collect_m[indices + 1],
-                        G_inv_collect_0[indices + 1],
-                        mu_arr,
+                        scaled_s_poly_coeffs[indicesp1],
+                        G_collect_m[indicesp1],
+                        K_collect_m[indicesp1],
+                        G_inv_mu_inv[indicesp1],
                     )
                     - _mathscr_v(
                         scaled_tau_arr_with_0[1:-1],
@@ -168,25 +197,24 @@ def _solve_for_coeffs(
                         scaled_s_poly_coeffs[indices],
                         G_collect_m[indices],
                         K_collect_m[indices],
-                        G_inv_collect_0[indices],
-                        mu_arr,
+                        G_inv_mu_inv[indices],
                     )
                 ).ravel(order='F')
+                
+            # Bottom boundary
+            RHS[-N:] -= _mathscr_v(
+                scaled_tau_arr_with_0[[-1]], 
+                np.array([0]),
+                Nscoeffs,
+                scaled_s_poly_coeffs[[-1], :],
+                G_collect_m[[-1], :N, :],
+                K_collect_m[[-1], :],
+                G_inv_mu_inv[[-1]],
+            ).ravel()
             
-            _mathscr_v_contribution_bottom = -_mathscr_v(
-                    scaled_tau_arr_with_0[[-1]], 
-                    np.array([0]),
-                    Nscoeffs,
-                    scaled_s_poly_coeffs[[-1], :],
-                    G_collect_m[[-1], :N, :],
-                    K_collect_m[[-1], :],
-                    G_inv_collect_0[[-1], :, :],
-                    mu_arr
-                ).ravel()
-            if NBDRF > 0:
-                _mathscr_v_contribution_bottom = (
-                    _mathscr_v_contribution_bottom
-                    + R
+            if there_is_BDRF_mode:
+                RHS[-N:] += (
+                    R
                     @ _mathscr_v(
                         scaled_tau_arr_with_0[[-1]],
                         np.array([0]),
@@ -194,126 +222,164 @@ def _solve_for_coeffs(
                         scaled_s_poly_coeffs[[-1], :],
                         G_collect_m[[-1], N:, :],
                         K_collect_m[[-1], :],
-                        G_inv_collect_0[[-1], :, :],
-                        mu_arr
+                        G_inv_mu_inv[[-1]],
                     ).ravel()
                 )
-                    
-            _mathscr_v_contribution = np.concatenate(
-                [
-                    _mathscr_v_contribution_top,
-                    _mathscr_v_contribution_middle,
-                    _mathscr_v_contribution_bottom,
-                ]
-            )   
-        else:
-            _mathscr_v_contribution = 0
         
         if there_is_beam_source:
-            RHS_middle = np.array([])
+            B_collect_m = B_collect[m, :, :]
+        
+            # Top boundary
+            RHS[:N] -= B_collect_m[0, N:]
+        
             if is_atmos_multilayered:
-                l_range = np.arange(1, NLayers)
-                RHS_middle = (
-                    (B_collect_m[l_range, :] - B_collect_m[l_range - 1, :])
-                    * np.exp(-scaled_tau_arr_with_0 / mu0)[l_range, None]
+                indicesp1 = np.arange(1, NLayers)
+                RHS[N:-N] += (
+                    (B_collect_m[indicesp1, :] - B_collect_m[indicesp1 - 1, :])
+                    * np.exp(-scaled_tau_arr_with_0 / mu0)[indicesp1, None]
                 ).ravel()
-                
+            
+            # Bottom boundary
             if there_is_BDRF_mode:
-                RHS = (
-                    np.concatenate(
-                        [
-                            b_neg_m - B_collect_m[0, N:],
-                            RHS_middle,
-                            b_pos_m
-                            + (mathscr_X_pos + R @ B_collect_m[-1, N:] - B_collect_m[-1, :N])
-                            * np.exp(-scaled_tau_arr_with_0[-1] / mu0),
-                        ]
-                    )
-                    + _mathscr_v_contribution
-                )
+                RHS[-N:] += (
+                    mathscr_X_pos + R @ B_collect_m[-1, N:] - B_collect_m[-1, :N]
+                ) * np.exp(-scaled_tau_arr_with_0[-1] / mu0)
+
             else:
-                RHS = (
-                    np.concatenate(
-                        [
-                            b_neg_m - B_collect_m[0, N:],
-                            RHS_middle,
-                            b_pos_m
-                            - B_collect_m[-1, :N] * np.exp(-scaled_tau_arr_with_0[-1] / mu0),
-                        ]
-                    )
-                    + _mathscr_v_contribution
-                )
-        else:
-            RHS_middle = np.zeros(NQuad * (NLayers - 1))
-            RHS = np.concatenate([b_neg_m, RHS_middle, b_pos_m]) + _mathscr_v_contribution
+                RHS[-N:] -= B_collect_m[-1, :N] * np.exp(-scaled_tau_arr_with_0[-1] / mu0)
+            
         # --------------------------------------------------------------------------------------------------------------------------
         
-        # Assemble LHS (much of this code is replicated in section 4 of the Comprehensive Documentation)
-        G_0_nn = G_collect_m[0, N:, :N]
-        G_0_np = G_collect_m[0, N:, N:]
-        G_L_pn = G_collect_m[-1, :N, :N]
-        G_L_nn = G_collect_m[-1, N:, :N]
-        G_L_pp = G_collect_m[-1, :N, N:]
-        G_L_np = G_collect_m[-1, N:, N:]
+        
+        # Assemble LHS and solve (much of this code is replicated in section 4 of the Comprehensive Documentation)
+        # --------------------------------------------------------------------------------------------------------------------------
+        G_0 = G_collect_m[0]
+        G_0_nn = G_0[N:, :N]
+        G_0_np = G_0[N:, N:]
+        E_01 = np.exp(K_collect_m[0, :N] * scaled_tau_arr_with_0[1])[None, :]
+        
+        G_L = G_collect_m[-1]
+        G_L_pn = G_L[:N, :N]
+        G_L_nn = G_L[N:, :N]
+        G_L_pp = G_L[:N, N:]
+        G_L_np = G_L[N:, N:]
         E_Lm1L = np.exp(
             K_collect_m[-1, :N] * (scaled_tau_arr_with_0[-1] - scaled_tau_arr_with_0[-2])
-        )
+        )[None, :]
         
-        LHS = np.zeros((dim, dim))
-        # BCs for the entire atmosphere
-        LHS[:N, :N] = G_0_nn
-        LHS[:N, N : NQuad] = (
-            G_0_np
-            * np.exp(K_collect_m[0, :N] * scaled_tau_arr_with_0[1])[None, :]
-        )
-        if there_is_BDRF_mode:
-            LHS[-N:, -NQuad : -N] = (G_L_pn - R @ G_L_nn) * E_Lm1L[None, :]
-            LHS[-N:, -N:] = G_L_pp - R @ G_L_np
-        else:
-            LHS[-N:, -NQuad : -N] = G_L_pn * E_Lm1L[None, :]
-            LHS[-N:, -N:] = G_L_pp
+        ################################## Use `sc.linalg.solve_banded` ##################################
+        if use_banded_solver:       
+        
+            # Directly assemble diagonal ordered form of LHS matrix
+            # ---------------- Top BC ----------------
+            G_0 = G_collect_m[0]
+            G_0_nn = G_0[N:, :N]     # (N,N)
+            G_0_np = G_0[N:, N:]     # (N,N)
 
-        # Interlayer / continuity BCs
-        for l in range(NLayers - 1):
-            G_l_pn = G_collect_m[l, :N, :N]
-            G_l_nn = G_collect_m[l, N:, :N]
-            G_l_ap = G_collect_m[l, :, N:]
-            G_lp1_an = G_collect_m[l + 1, :, :N]
-            G_lp1_pp = G_collect_m[l + 1, :N, N:]
-            G_lp1_np = G_collect_m[l + 1, N:, N:]
-            scaled_tau_arr_lm1 = scaled_tau_arr_with_0[l]
-            scaled_tau_arr_l = scaled_tau_arr_with_0[l + 1]
-            scaled_tau_arr_lp1 = scaled_tau_arr_with_0[l + 2]
-            # Postive eigenvalues
-            K_l_pos = K_collect_m[l, N:]
-            K_lp1_pos = K_collect_m[l + 1, N:]
-            E_lm1l = np.exp(K_l_pos * (scaled_tau_arr_lm1 - scaled_tau_arr_l))
-            E_llp1 = np.exp(K_lp1_pos * (scaled_tau_arr_l - scaled_tau_arr_lp1))
-            
-            start_row = N + l * NQuad
-            start_col = l * NQuad
-            LHS[start_row : N + start_row, start_col : N + start_col] = G_l_pn * E_lm1l[None, :]
-            LHS[N + start_row : 2 * N + start_row, start_col : N + start_col] = G_l_nn * E_lm1l[None, :]
-            LHS[start_row : 2 * N + start_row, N + start_col : 2 * N + start_col] = G_l_ap
-            LHS[start_row : 2 * N + start_row, 2 * N + start_col : 3 * N + start_col] = -G_lp1_an
-            LHS[start_row : N + start_row, 3 * N + start_col : 4 * N + start_col] = -G_lp1_pp * E_llp1[None, :]
-            LHS[N + start_row : 2 * N + start_row, 3 * N + start_col : 4 * N + start_col] = -G_lp1_np * E_llp1[None, :]
-            
-        # --------------------------------------------------------------------------------------------------------------------------
+            _view_slanted(0, 0, N)[:] = G_0_nn  # `_view_slanted` was defined with respect to `LHS_dof`
+            _view_slanted(0, N, N)[:] = G_0_np * E_01
+
+            # ---------------- Bottom BC ----------------
+            if there_is_BDRF_mode:
+                _view_slanted(i_bot, j_bot_left, N)[:] = (G_L_pn - R @ G_L_nn) * E_Lm1L
+                _view_slanted(i_bot, i_bot, N)[:] = G_L_pp - R @ G_L_np
+            else:
+                _view_slanted(i_bot, j_bot_left, N)[:] = G_L_pn * E_Lm1L
+                _view_slanted(i_bot, i_bot, N)[:] = G_L_pp
+
+            # ---------------- Interlayer / continuity BCs ----------------
+            for l in range(NLayers - 1):
+                G_l = G_collect_m[l]
+                G_l_pn = G_l[:N, :N]      # (N,N)
+                G_l_nn = G_l[N:, :N]      # (N,N)
+                G_l_ap = G_l[:, N:]       # (2N,N)
+
+                G_lp1 = G_collect_m[l + 1]
+                G_lp1_an = G_lp1[:, :N]   # (2N,N)
+                G_lp1_pp = G_lp1[:N, N:]  # (N,N)
+                G_lp1_np = G_lp1[N:, N:]  # (N,N)
+
+                tau_lm1 = scaled_tau_arr_with_0[l]
+                tau_l   = scaled_tau_arr_with_0[l + 1]
+                tau_lp1 = scaled_tau_arr_with_0[l + 2]
+
+                E_lm1l = np.exp(K_collect_m[l,     N:] * (tau_lm1 - tau_l))[None, :]    # (N,)
+                E_llp1 = np.exp(K_collect_m[l + 1, N:] * (tau_l   - tau_lp1))[None, :]  # (N,)
+
+                start_row = N + l * NQuad
+                start_col = l * NQuad
+
+                _view_slanted(start_row,     start_col,     N)[:] = G_l_pn * E_lm1l
+                _view_slanted(start_row + N, start_col,     N)[:] = G_l_nn * E_lm1l
+                _view_slanted(start_row,     start_col + N, NQuad)[:] = G_l_ap
+
+                _view_slanted(start_row,     start_col + NQuad, NQuad)[:] = -G_lp1_an
+                _view_slanted(start_row,     start_col + 3 * N, N)[:] = -G_lp1_pp * E_llp1
+                _view_slanted(start_row + N, start_col + 3 * N, N)[:] = -G_lp1_np * E_llp1
         
-        # Solve the system
-        if use_banded_solver:
+            # Solve the system
             C_m = sc.linalg.solve_banded(
                 (Nsupsubdiags, Nsupsubdiags),
-                to_diag_ordered_form(LHS, Nsupsubdiags, Nsupsubdiags),
+                LHS_dof,
                 RHS,
-                True,
+                False,
                 True,
                 False,
             )
-        else:
-            C_m = np.linalg.solve(LHS, RHS)
-
-        GC_collect[m, :, :, :] = G_collect_m * C_m.reshape(NLayers, NQuad)[:, None, :]
         
-    return GC_collect
\ No newline at end of file
+        ##################################################################################################
+        ##################################### Use `np.linalg.solve` ######################################
+        else:
+        
+            # ---------------- Top BC ----------------
+            LHS[:N, :N] = G_0_nn
+            LHS[:N, N : NQuad] = G_0_np * E_01
+            
+            # ---------------- Bottom BC ----------------
+            if there_is_BDRF_mode:
+                LHS[-N:, -NQuad : -N] = (G_L_pn - R @ G_L_nn) * E_Lm1L
+                LHS[-N:, -N:] = G_L_pp - R @ G_L_np
+            else:
+                LHS[-N:, -NQuad : -N] = G_L_pn * E_Lm1L
+                LHS[-N:, -N:] = G_L_pp
+
+            # ---------------- Interlayer / continuity BCs ----------------
+            for l in range(NLayers - 1):
+                G_l = G_collect_m[l]
+                G_l_pn = G_l[:N, :N]
+                G_l_nn = G_l[N:, :N]
+                G_l_ap = G_l[:, N:]
+                
+                G_lp1 = G_collect_m[l + 1]
+                G_lp1_an = G_lp1[:, :N]
+                G_lp1_pp = G_lp1[:N, N:]
+                G_lp1_np = G_lp1[N:, N:]
+                
+                scaled_tau_arr_lm1 = scaled_tau_arr_with_0[l]
+                scaled_tau_arr_l = scaled_tau_arr_with_0[l + 1]
+                scaled_tau_arr_lp1 = scaled_tau_arr_with_0[l + 2]
+                
+                # Postive eigenvalues
+                E_lm1l = np.exp(K_collect_m[l, N:] * (scaled_tau_arr_lm1 - scaled_tau_arr_l))
+                E_llp1 = np.exp(K_collect_m[l + 1, N:] * (scaled_tau_arr_l - scaled_tau_arr_lp1))
+                
+                start_row = N + l * NQuad
+                start_col = l * NQuad
+                
+                LHS[start_row : N + start_row, start_col : N + start_col] = G_l_pn * E_lm1l
+                LHS[N + start_row : NQuad + start_row, start_col : N + start_col] = G_l_nn * E_lm1l
+                LHS[start_row : NQuad + start_row, N + start_col : NQuad + start_col] = G_l_ap
+                
+                LHS[start_row : NQuad + start_row, NQuad + start_col : 3 * N + start_col] = -G_lp1_an
+                LHS[start_row : N + start_row, 3 * N + start_col : 4 * N + start_col] = -G_lp1_pp * E_llp1
+                LHS[N + start_row : NQuad + start_row, 3 * N + start_col : 4 * N + start_col] = -G_lp1_np * E_llp1
+                
+            # Solve the system
+            C_m = np.linalg.solve(LHS, RHS)
+                
+            ################################################################################################## 
+            # --------------------------------------------------------------------------------------------------------------------------
+        
+        C_collect[m, :] = C_m
+        
+    return G_collect * C_collect.reshape((NFourier, NLayers, 1, NQuad))
