@@ -1,6 +1,32 @@
 import numpy as np
 import scipy as sc
 from math import pi
+import warnings
+
+
+def prepend(arr, arr_len, value):
+    """Prepend value to array. 
+    Faster than ``np.insert(arr, 0, value)``.
+
+    Parameters
+    ----------
+    arr : array
+        The 1D array to which to prepend.
+    arr_len : int
+        Length of original array (``arr``).
+    value : float
+        Value to prepend.
+
+    Returns
+    -------
+    array
+        Array of length ``arr_len + 1`` with ``value`` prepended.
+
+    """
+    new_arr = np.empty(arr_len + 1)
+    new_arr[1:] = arr
+    new_arr[0] = value
+    return new_arr
 
 
 
@@ -12,13 +38,13 @@ def transform_interval(arr, c, d, a, b):
     arr : array
         The 1D array to transform.
     c : float
-        The beginning of interval [c, d].
+        Lower bound of interval [c, d].
     d : float
-        The end of interval [c, d].
+        Upper bound of interval [c, d].
     a : float, optional
-        The beginning of interval [a, b].
+        Lower bound of interval [a, b].
     b : float, optional
-        The end of interval [a, b].
+        Upper bound of interval [a, b].
 
     Returns
     -------
@@ -38,13 +64,13 @@ def transform_weights(weights, c, d, a, b):
     weights : array
         The weights to transform.
     c : float
-        The beginning of interval [c, d].
+        Lower bound of interval [c, d].
     d : float
-        The end of interval [c, d].
+        Upper bound of interval [c, d].
     a : float, optional
-        The beginning of interval [a, b].
+        Lower bound of interval [a, b].
     b : float, optional
-        The end of interval [a, b].
+        Upper bound of interval [a, b].
 
     Returns
     -------
@@ -95,9 +121,9 @@ def Gauss_Legendre_quad(N, c=0, d=1):
     N : int (will be converted to int)
         Number of quadrature nodes.
     c : float, optional
-        Start of integration interval.
+        Lower bound of integration interval.
     d : float, optional
-        End of integration interval.
+        Upper bound of integration interval.
 
     Returns
     -------
@@ -141,7 +167,7 @@ def Clenshaw_Curtis_quad(Nphi, c=0, d=(2 * pi)):
     Nphi_pos = Nphi // 2
     phi_arr_pos = np.cos(pi * np.arange(Nphi_pos) / Nphi)
     phi_arr = np.concatenate([-phi_arr_pos, [0], np.flip(phi_arr_pos)])
-    diff = np.insert(2 / (1 - 4 * np.arange(1, Nphi_pos + 1) ** 2), 0, 2)
+    diff = prepend(2 / (1 - 4 * np.arange(1, Nphi_pos + 1) ** 2), Nphi_pos, 2)
     weights_phi_pos = sc.fft.idct(diff, type=1)
     weights_phi_pos[0] /= 2
     full_weights_phi = np.hstack((weights_phi_pos, np.flip(weights_phi_pos[:-1])))
@@ -248,18 +274,18 @@ def generate_diff_act_flux_funcs(u0):
         Actinic flux function with argument ``tau`` (type: array) for positive (upward) ``mu`` values.
         Returns the diffuse flux magnitudes (type: array).
         Pass ``is_antiderivative_wrt_tau = True`` (defaults to ``False``)
-        to switch to an antiderivative of the function with respect to ``tau``.
+        to switch to an antiderivative of this function with respect to ``tau``.
         Pass ``return_tau_arr`` to return ``tau_arr`` (defaults to ``False``).
     Fm_act(tau) : function
         Actinic flux function with argument ``tau`` (type: array) for negative (downward) ``mu`` values.
         Returns the diffuse flux magnitudes (type: array).
         Pass ``is_antiderivative_wrt_tau = True`` (defaults to ``False``)
-        to switch to an antiderivative of the function with respect to ``tau``.
+        to switch to an antiderivative of this function with respect to ``tau``.
         Pass ``return_tau_arr`` to return ``tau_arr`` (defaults to ``False``).
 
     """
     N = len(u0(0)) // 2
-    GL_weights = Gauss_Legendre_quad(N, 0, 1)[1]
+    GL_weights = Gauss_Legendre_quad(N)[1]
 
     # Note that the zeroth axis of the array u0(tau) captures variation with mu
     def flux_act_up(tau, is_antiderivative_wrt_tau=False, return_tau_arr=False):
@@ -314,11 +340,12 @@ def Planck(T, WVNM):
     not_zeros_ind = T != 0
     
     results = np.zeros(len(T))
-    results[~not_zeros_ind] = 0
-    
-    # Coded in this way to prevent overflow
-    expterm = np.exp(-100 * sc.constants.h * sc.constants.c * WVNM / (sc.constants.k * T[not_zeros_ind]))
-    results[not_zeros_ind] = (2e8 * sc.constants.h * sc.constants.c**2 * WVNM**3 * expterm) / (1 - expterm)
+    if np.sum(not_zeros_ind) > 0:
+        results[~not_zeros_ind] = 0
+        
+        # Coded in this way to prevent overflow
+        expterm = np.exp(-100 * sc.constants.h * sc.constants.c * WVNM / (sc.constants.k * T[not_zeros_ind]))
+        results[not_zeros_ind] = (2e8 * sc.constants.h * sc.constants.c**2 * WVNM**3 * expterm) / (1 - expterm)
         
     return np.squeeze(results)[()]
     
@@ -386,8 +413,8 @@ def linear_spline_coefficients(x, y, check_inputs=True):
 def generate_s_poly_coeffs(tau_arr, TEMPER, WVNMLO, WVNMHI, **kwargs):
     """Generate DISORT-equivalent ``s_poly_coeffs`` for input into ``pydisort``.
     This convenience function is provided to help match the inputs for Stamnes' DISORT to those for PythonicDISORT.
-    Users will have to manually adjust emissivities, however, and they should note that PythonicDISORT allows much
-    more flexibility in choosing blackbody emission profiles than DISORT.
+    Note that the coefficients will be multiplied by emissivity factors equal to ``1 - omega_arr`` inside
+    `pydisort`, i.e. Kirchoff's law of thermal radiation is enforced, just like in Stamnes' DISORT. 
     
     Parameters
     ----------
@@ -413,12 +440,19 @@ def generate_s_poly_coeffs(tau_arr, TEMPER, WVNMLO, WVNMHI, **kwargs):
     """
     tau_arr = np.atleast_1d(tau_arr)
     if not len(TEMPER) == len(tau_arr) + 1:
-        raise ValueError("Missing temperature specification at some boundaries / interfaces.")
-    
-    tau_arr_with_0 = np.insert(tau_arr, 0, 0)
-    blackbody_emission_at_each_boundary = sc.integrate.quad_vec(lambda WVNM: Planck(TEMPER, WVNM), WVNMLO, WVNMHI, **kwargs)[0]
-    
-    return linear_spline_coefficients(tau_arr_with_0, blackbody_emission_at_each_boundary, check_inputs=False)
+        raise ValueError(
+            "Missing temperature specification at some boundaries / interfaces."
+        )
+
+    tau_arr_with_0 = prepend(tau_arr, len(tau_arr), 0)
+    blackbody_emission_at_each_boundary = sc.integrate.quad_vec(
+        lambda WVNM: Planck(TEMPER, WVNM), WVNMLO, WVNMHI, **kwargs
+    )[0]
+
+    return linear_spline_coefficients(
+        tau_arr_with_0, blackbody_emission_at_each_boundary, check_inputs=False
+    )
+
 
 
 
@@ -432,34 +466,43 @@ def generate_emissivity_from_BDRF(N, zeroth_BDRF_Fourier_mode):
     ----------
     N : int
         Number of upper hemisphere quadrature nodes. Equal to ``NQuad // 2``.
-    zeroth_BDRF_Fourier_mode : function
+    zeroth_BDRF_Fourier_mode : function or float
         Zeroth BDRF Fourier mode with arguments ``mu, -mu_p`` of type array
         and which output has the same dimensions as the outer product of the two arrays.
+        A scalar input represents a constant function and in that case the output 
+        is simply one minus the scalar input.
 
     Returns
     -------
-    array
+    array or float
         Emissivity for the blackbody contribution to the lower boundary source ``b_neg``.
     """
-    mu_arr_pos, W = Gauss_Legendre_quad(N)
-    return 1 - 2 * zeroth_BDRF_Fourier_mode(mu_arr_pos, mu_arr_pos) * mu_arr_pos[None, :] @ W
+    if np.isscalar(zeroth_BDRF_Fourier_mode):
+        return 1 - zeroth_BDRF_Fourier_mode
+    else:
+        mu_arr_pos, W = Gauss_Legendre_quad(N)
+        return (
+            1 - 2 * zeroth_BDRF_Fourier_mode(mu_arr_pos, mu_arr_pos) * mu_arr_pos[None, :] @ W
+        )
 
 
 
-def cache_BDRF_Fourier_modes(N, mu0, BDRF_Fourier_modes):
-    """If the same BDRF, number of streams and ``mu0`` will be used repeatedly,
-    consider using this function to cache ``BDRF_Fourier_modes`` and instead input
-    ``cached_BDRF_Fourier_modes`` into ``pydisort``.
+def cache_BDRF_Fourier_modes(N, BDRF_Fourier_modes, mu0=0):
+    """If the same BDRF and number of streams will be used repeatedly,
+    consider using this function to cache ``BDRF_Fourier_modes`` and pass
+    ``cached_BDRF_Fourier_modes`` to ``pydisort``. It is also possible
+    to cache ``BDRF_Fourier_modes`` with respect to a fixed ``mu0``. 
 
     Parameters
     ----------
     N : int
         Number of upper hemisphere quadrature nodes. Equal to ``NQuad // 2``.
-    mu0 : float
-        Cosine of polar angle of the incident beam.
-    BDRF_Fourier_modes : list of functions
-        BDRF Fourier modes, each a scalar, or a function with arguments ``mu, -mu_p`` of type array
+    BDRF_Fourier_modes : list of functions and scalars
+        BDRF Fourier modes, each a scalar representing a constant function, 
+        or a function with arguments ``mu, -mu_p`` of type array
         which output has the same dimensions as the outer product of the two arrays.
+    mu0 : float, optional
+        Cosine of polar angle of the incident beam.
 
     Returns
     -------
@@ -468,30 +511,61 @@ def cache_BDRF_Fourier_modes(N, mu0, BDRF_Fourier_modes):
         and which output is a scalar (if the Fourier mode was a scalar) 
         or has the same dimensions as the outer product of the two arrays.
     """
+    
+    if (0 < mu0 and mu0 <= 1):
+        mu0_caching = True
+    else:
+        mu0_caching = False
+        warnings.warn("No caching with respect to `mu0`.")
+    
     NBDRF = len(BDRF_Fourier_modes)
     mu_arr_pos = Gauss_Legendre_quad(N)[0]
-
-    BDRF_Fourier_modes_evaluated = [
-        (
-            None
-            if np.isscalar(BDRF_Fourier_modes[m])
-            else BDRF_Fourier_modes[m](mu_arr_pos, np.append(mu_arr_pos, mu0))
-        )
-        for m in range(NBDRF)
-    ]
-
-    cached_BDRF_Fourier_modes = [
-        (
-            lambda mu, neg_mup, m=m: BDRF_Fourier_modes[m]
-            if np.isscalar(BDRF_Fourier_modes[m])
-            else (
-                BDRF_Fourier_modes_evaluated[m][:, [-1]]
-                if len(neg_mup) == 1
-                else BDRF_Fourier_modes_evaluated[m][:, :-1]
+    
+    if mu0_caching:
+        BDRF_Fourier_modes_evaluated = [
+            (
+                BDRF_Fourier_modes[m]
+                if np.isscalar(BDRF_Fourier_modes[m])
+                else BDRF_Fourier_modes[m](mu_arr_pos, np.append(mu_arr_pos, mu0))
             )
-        )
-        for m in range(NBDRF)
-    ]
+            for m in range(NBDRF)
+        ]
+
+        cached_BDRF_Fourier_modes = [
+            (
+                lambda mu, neg_mup, m=m: BDRF_Fourier_modes_evaluated[m]
+                if np.isscalar(BDRF_Fourier_modes_evaluated[m])
+                else (
+                    BDRF_Fourier_modes_evaluated[m][:, [-1]]
+                    if len(neg_mup) == 1
+                    else BDRF_Fourier_modes_evaluated[m][:, :-1]
+                )
+            )
+            for m in range(NBDRF)
+        ]
+        
+    else:
+        BDRF_Fourier_modes_evaluated = [
+            (
+                BDRF_Fourier_modes[m]
+                if np.isscalar(BDRF_Fourier_modes[m])
+                else BDRF_Fourier_modes[m](mu_arr_pos, mu_arr_pos)
+            )
+            for m in range(NBDRF)
+        ]
+
+        cached_BDRF_Fourier_modes = [
+            (
+                lambda mu, neg_mup, m=m: BDRF_Fourier_modes_evaluated[m]
+                if np.isscalar(BDRF_Fourier_modes_evaluated[m])
+                else (
+                    BDRF_Fourier_modes[m](mu_arr_pos, neg_mup)
+                    if len(neg_mup) == 1
+                    else BDRF_Fourier_modes_evaluated[m]
+                )
+            )
+            for m in range(NBDRF)
+        ]
 
     return cached_BDRF_Fourier_modes
 
@@ -551,22 +625,24 @@ def interpolate(u):
     -------
     u_interpol : function
         Intensity function with arguments ``(mu, tau, phi)`` (for ``u``) 
-        or ``(mu, tau)`` (for ``u0``) each of type array or float.
+        or ``(mu, tau)`` (for ``u0``) each an array or scalar.
         Returns an ndarray with axes corresponding to variation with each argument in the same order.
         Pass ``is_antiderivative_wrt_tau = True`` (defaults to ``False``)
-        to switch to an antiderivative of the function with respect to ``tau``.
+        to switch to an antiderivative of this function with respect to ``tau``.
         Pass ``return_Fourier_error = True`` (defaults to ``False``) to return the 
         Cauchy / Fourier convergence evaluation (type: float) for the last Fourier term.
         Pass ``return_tau_arr`` to return ``tau_arr`` (defaults to ``False``).
 
     """
+    
     N = len(u(0, 0)) // 2
     mu_arr_pos = Gauss_Legendre_quad(N)[0]
     
     u_pos_interpol = sc.interpolate.BarycentricInterpolator(mu_arr_pos)
     u_neg_interpol = sc.interpolate.BarycentricInterpolator(-mu_arr_pos)
     
-    if u.__code__.co_argcount == 5: # Function is u instead of u0
+    argcount = u.__code__.co_argcount
+    if argcount == 5: # Function is `u` instead of `u0`
         def u_interpol(mu, tau, phi, is_antiderivative_wrt_tau=False, return_Fourier_error=False, return_tau_arr=False):
             if not np.all(np.abs(mu) <= 1):
                 raise ValueError("mu values must be between -1 and 1.")
@@ -595,7 +671,7 @@ def interpolate(u):
             else:
                 return np.squeeze(results)[()]
     
-    elif u.__code__.co_argcount == 4: # Function is u0 instead of u
+    elif argcount == 4: # Function is u0 instead of u
         def u_interpol(mu, tau, is_antiderivative_wrt_tau=False, return_Fourier_error=False, return_tau_arr=False):
             if not np.all(np.abs(mu) <= 1):
                 raise ValueError("mu values must be between -1 and 1.")
@@ -624,7 +700,7 @@ def interpolate(u):
             else:
                 return np.squeeze(results)[()]
     else:
-        raise ValueError("This subroutine can only interpolate u or u0")
+        raise ValueError("This subroutine can only interpolate u or u0.")
     
     return u_interpol
          
@@ -667,17 +743,16 @@ def to_diag_ordered_form(A, Nsuperdiags, Nsubdiags):
 
 
     
-def _mathscr_v(tau,                              # Input optical depths
-                l,                               # Layer index of each input optical depth
-                Nscoeffs,                        # Number of isotropic source polynomial coefficients
-                s_poly_coeffs,                   # Polynomial coefficients of isotropic source
-                G,                               # Eigenvector matrices
-                K,                               # Eigenvalues
-                G_inv,                           # Inverse of eigenvector matrix
-                mu_arr,                          # Quadrature nodes for both hemispheres
-                is_antiderivative_wrt_tau=False, # Switch to an antiderivative of the function?
-                autograd_compatible=False,       # Should the output functions be compatible with autograd?
-                ):
+def _mathscr_v(tau,                             # Input optical depths
+               l,                               # Layer index of each input optical depth
+               Nscoeffs,                        # Number of isotropic source polynomial coefficients
+               s_poly_coeffs,                   # Polynomial coefficients of isotropic source
+               G,                               # Eigenvector matrices
+               K,                               # Eigenvalues
+               G_inv_mu_inv,                    # Inverse of eigenvector matrix @ (1 / `mu_arr`)
+               is_antiderivative_wrt_tau=False, # Switch to an antiderivative of the function?
+               autograd_compatible=False,       # Should the output functions be compatible with autograd?
+               ):
     """Particular solution for isotropic internal sources.
     Refer to section 3.6.1 of the Comprehensive Documentation.
     It has many seemingly redundant arguments to maximize 
@@ -690,102 +765,118 @@ def _mathscr_v(tau,                              # Input optical depths
     | ``tau``                       | ``Ntau``                                  |
     | ``l``                         | ``Ntau``                                  |
     | ``Nscoeffs``                  | scalar                                    |
-    | ``s_poly_coeffs``             | ``NLayers x Nscoeffs``                    |
-    | ``G``                         | ``NLayers<= x NQuad x NQuad``             |
-    | ``K``                         | ``NLayers<= x NQuad``                     |
-    | ``G_inv``                     | ``NLayers<= x NQuad x NQuad`` or ``None`` |
-    | ``mu_arr``                    | ``NQuad``                                 |
+    | ``s_poly_coeffs``             | ``NL x Nscoeffs``                         |
+    | ``G``                         | ``NL x NQuad x NQuad``                    |
+    | ``K``                         | ``NL x NQuad``                            |
+    | ``G_inv_mu_inv``              | ``NL x NQuad``                            |
     | ``is_antiderivative_wrt_tau`` | boolean                                   |
     | ``autograd_compatible``       | boolean                                   |
     
-    Notable internal variables of _mathscr_v
-    |     Variable     |                Shape                  | 
-    | ---------------- | ------------------------------------- |
-    | i_arr            | ``Nscoeffs``                          |
-    | i_arr_repeat     | ``Nscoeffs*(Nscoeffs+1)/2``           | Using triangular number formula
-    | j_arr            | ``Nscoeffs*(Nscoeffs+1)/2``           | Using triangular number formula
-    | s_poly_coeffs_nj | ``NLayers x Nscoeffs*(Nscoeffs+1)/2`` |
-    | OUTPUT           | ``NQuad x Ntau``                      |
     """
-    n = Nscoeffs - 1
-    
     if autograd_compatible:
         import autograd.numpy as np
-    
-        def mathscr_b(i):
-            """
-            Notable internal variables of mathscr_b
-            |     Variable     |                  Shape                  |
-            | ---------------- | --------------------------------------- |
-            | j_arr            | ``i + 1``                               |
-            | s_poly_coeffs_nj | ``i + 1``                               |
-            | OUTPUT           | ``Nscoeffs x (i + 1) x NQuad``          |
-            """
-        
-            j_arr = np.arange(i + 1)
-            s_poly_coeffs_nj = s_poly_coeffs[:, n - j_arr]
-            return np.sum(
-                (sc.special.factorial(n - j_arr) / sc.special.factorial(n - i))[None, None, :]
-                * K[:, :, None] ** -(i - j_arr + 1)[None, None, :]
-                * s_poly_coeffs_nj[:, None, :],
-                axis=-1,
-            )
-        mathscr_v_coeffs = np.array(list(map(mathscr_b, range(Nscoeffs))))
     else:
         import numpy as np
-        
-        shape = np.shape(K)
-        i_arr = np.arange(Nscoeffs)
-        i_arr_repeat = np.repeat(i_arr, i_arr + 1)
-        j_arr = np.concatenate([i_arr[:i] for i in range(1, Nscoeffs + 1)])
-        s_poly_coeffs_nj = s_poly_coeffs[:, n - j_arr]
 
-        mathscr_v_coeffs = np.zeros((Nscoeffs, shape[0], shape[1]))
-        np.add.at(
-            mathscr_v_coeffs,
-            i_arr_repeat,
-            np.moveaxis(
-                (sc.special.factorial(n - j_arr) / sc.special.factorial(n - i_arr_repeat))[
-                    None, None, :
-                ]
-                * K[:, :, None] ** -(i_arr_repeat - j_arr + 1)[None, None, :]
-                * s_poly_coeffs_nj[:, None, :],
-                -1,
-                0,
-            ),
-        )
-    
-    powers = np.arange(Nscoeffs - 1, -1, -1)[None, :]
-    if is_antiderivative_wrt_tau:
-        tau_poly = tau[:, None] ** (powers + 1) / (powers + 1)
+    K_inv = 1 / K  # (NL, NQuad)
+
+    # --------------------- Nscoeffs == 1 ---------------------
+    if Nscoeffs == 1:
+        s0 = s_poly_coeffs[:, 0:1]
+        c0w = s0 * K_inv * G_inv_mu_inv  # (NL, 1) * (NL, NQuad) -> (NL, NQuad)
+
+        if is_antiderivative_wrt_tau:
+            p0 = tau                 
+        else:
+            p0 = np.ones_like(tau)       
+            
+        d0 = np.einsum("lij, lj -> il", G, c0w, optimize=True)
+
+        return d0[:, l] * p0[None, :]
+
+    # --------------------- Nscoeffs == 2 ---------------------
+    elif Nscoeffs == 2:
+        s1 = s_poly_coeffs[:, 1:2]  # (NL, )
+        s0 = s_poly_coeffs[:, 0:1]  # (NL, )
+
+        K_inv2 = K_inv * K_inv
+
+        c0w = s1 * K_inv * G_inv_mu_inv
+        c1w = (s0 * K_inv + s1 * K_inv2) * G_inv_mu_inv
+
+        if is_antiderivative_wrt_tau:
+            p0 = tau * tau / 2
+            p1 = tau
+        else:
+            p0 = tau
+            p1 = np.ones_like(tau) 
+
+        d0 = np.einsum("lij, lj -> il", G, c0w, optimize=True)  # (NQuad, NL)
+        d1 = np.einsum("lij, lj -> il", G, c1w, optimize=True)  # (NQuad, NL)
+
+        return d0[:, l] * p0[None, :] + d1[:, l] * p1[None, :]
+
+    # --------------------- Nscoeffs > 2 ----------------------
     else:
-        tau_poly = tau[:, None] ** powers
-    
-    return np.einsum(
-        "tik, tc, ctk -> it",
-        G[l, :, :],
-        tau_poly,
-        (mathscr_v_coeffs * (G_inv @ (1 / mu_arr))[None, :, :])[:, l, :],
-        optimize=True,
-    )
-    
+        if Nscoeffs > 10:
+            warnings.warn("`Nscoeffs` is large and may cause instability.")
+        n = Nscoeffs - 1
+
+        # K_invP[..., powers] = K_inv^(powers + 1); (NL, NQuad, Nscoeffs)
+        K_invP = np.cumprod(
+            np.broadcast_to(K_inv[:, :, None], K_inv.shape + (Nscoeffs,)),
+            axis=-1
+        )
+
+        ii = np.arange(Nscoeffs, dtype=int)[:, None]
+        pp = np.arange(Nscoeffs, dtype=int)[None, :]
+        jj = ii - pp
+
+        # Factorials
+        fac = np.empty(Nscoeffs)
+        fac[0] = 1
+        if Nscoeffs > 1:
+            fac[1:] = np.cumprod(np.arange(1, Nscoeffs))
+        fac_rev = fac[::-1]
+        inv_fac_rev = 1 / fac_rev
+
+        weighted_mathscr_a = s_poly_coeffs[:, ::-1] * fac_rev[None, :]
+        
+        # Only take lower triangle values
+        lower_tri = np.where((jj >= 0)[None, :, :], np.take(weighted_mathscr_a, jj, axis=1), 0)
+        
+        unweighted_mathscr_b_sigma = np.einsum("lkp, lip -> lki", K_invP, lower_tri, optimize=True)
+        mathscr_b_sigma = (unweighted_mathscr_b_sigma * inv_fac_rev[None, None, :]).transpose(2, 0, 1)
+        mathscr_b_right = mathscr_b_sigma * G_inv_mu_inv[None, :, :]
+        mathscr_b = np.einsum("lqk, plk- > lqp", G, mathscr_b_right, optimize=True)
+
+        powers = np.arange(Nscoeffs - 1, -1, -1)[None, :]
+        if is_antiderivative_wrt_tau:
+            p = powers + 1
+            tau_poly = (tau[:, None] ** p) / p
+        else:
+            tau_poly = tau[:, None] ** powers
+
+        return np.einsum("tqp, tp -> qt", mathscr_b[l, :, :], tau_poly, optimize=True)
+
 
 
 def _compare(results, mu_to_compare, reorder_mu, flux_up, flux_down, u=None):
     """Performs pointwise comparisons between results from Stamnes' DISORT,
     which are stored in ``.npz`` files, against results from PythonicDISORT. Used in our PyTests.
-
+    See the ``*_test`` Jupyter Notebooks to see how this function is used and the arguments that go into it.
+    
     """
     # Load saved results from Stamnes' DISORT
     flup = results["flup"]
     rfldn = results["rfldn"]
     rfldir = results["rfldir"]
-    if u != None:
+    if u is not None:
         uu = results["uu"]
 
     # Load comparison points
     tau_test_arr = results["tau_test_arr"]
-    if u != None:
+    if u is not None:
         phi_arr = results["phi_arr"]
 
     # Perform and print the comparisons
@@ -800,7 +891,7 @@ def _compare(results, mu_to_compare, reorder_mu, flux_up, flux_down, u=None):
         diff_flux_up,
         flup,
         out=np.zeros_like(diff_flux_up),
-        where=flup > 1e-8,
+        where=(flup != 0),
     )
     print("Difference =", np.max(diff_flux_up))
     print("Difference ratio =", np.max(ratio_flux_up))
@@ -813,7 +904,7 @@ def _compare(results, mu_to_compare, reorder_mu, flux_up, flux_down, u=None):
         diff_flux_down_diffuse,
         rfldn,
         out=np.zeros_like(diff_flux_down_diffuse),
-        where=rfldn > 1e-8,
+        where=(rfldn != 0),
     )
     print("Difference =", np.max(diff_flux_down_diffuse))
     print(
@@ -829,7 +920,7 @@ def _compare(results, mu_to_compare, reorder_mu, flux_up, flux_down, u=None):
         diff_flux_down_direct,
         rfldir,
         out=np.zeros_like(diff_flux_down_direct),
-        where=rfldir > 1e-8,
+        where=(rfldir != 0),
     )
     print("Difference =", np.max(diff_flux_down_direct))
     print(
@@ -838,14 +929,15 @@ def _compare(results, mu_to_compare, reorder_mu, flux_up, flux_down, u=None):
     )
     print()
     
-    if u != None:
+    if u is not None:
         # Intensity
-        diff = np.abs(uu - u(tau_test_arr, phi_arr)[reorder_mu])[mu_to_compare]
+        u_cache = u(tau_test_arr, phi_arr)[reorder_mu].reshape(np.shape(uu))
+        diff = np.abs(uu - u_cache)[mu_to_compare]
         diff_ratio = np.divide(
             diff,
             uu[mu_to_compare],
             out=np.zeros_like(diff),
-            where=uu[mu_to_compare] > 1e-8,
+            where=(uu[mu_to_compare] != 0),
         )
         max_diff_tau_index = np.argmax(np.max(np.max(diff, axis=0), axis=1))
         max_ratio_tau_index = np.argmax(np.max(np.max(diff_ratio, axis=0), axis=1))
